#include <lmic.h>
#include <hal/hal.h>
#include <SPI.h>
#include "credentials.h"
#include "pinMapping.h"

#include <Adafruit_Sensor.h>
#include <Adafruit_BME280.h>
#include "SSD1306.h"

#define I2C_SDA 21
#define I2C_SCL 22

#define SEALEVELPRESSURE_HPA (1032.75)
Adafruit_BME280 bme; // I2C
unsigned long delayTime;
bool bBme280isConnected;

SSD1306  display(0x3c, I2C_SDA, I2C_SCL);

#define BUTTON_PRG 0
//#define BUTTON_PRG 23

#define uS_TO_S_FACTOR 1000000  /* Conversion factor for micro seconds to seconds */
#define TIME_TO_SLEEP  10*60       /* Time ESP32 will go to sleep (in seconds) */
//#define TIME_TO_SLEEP  15       /* Time ESP32 will go to sleep (in seconds) */
#define TIME_TO_REBOOT  3       /* Time ESP32 will go to sleep (in seconds) */

RTC_DATA_ATTR int bootCount = 0;

// -----------------------------------------------------------------------------
// Globals
// -----------------------------------------------------------------------------
#define ESP32

#ifndef CFG_eu868
#error "This script is meant to connect to TTN EU network at 868MHz"
#endif

#define TX_INTERVAL         300  /* Time between transmissions. Using Deep Sleep is useless*/

// These callbacks are only used in over-the-air activation, so they are
// left empty here (we cannot leave them out completely unless
// DISABLE_JOIN is set in config.h, otherwise the linker will complain).
void os_getArtEui (u1_t* buf) { }
void os_getDevEui (u1_t* buf) { }
void os_getDevKey (u1_t* buf) { }

// Job
static osjob_t sendjob;

// Value
unsigned long autoincrement = 0;

// -----------------------------------------------------------------------------
// Deep Sleep {
// -----------------------------------------------------------------------------
void print_wakeup_reason() {
  esp_sleep_wakeup_cause_t wakeup_reason;

  wakeup_reason = esp_sleep_get_wakeup_cause();

  switch (wakeup_reason)
  {
    case 1  : Serial.println("Wakeup caused by external signal using RTC_IO"); break;
    case 2  : Serial.println("Wakeup caused by external signal using RTC_CNTL"); break;
    case 3  : Serial.println("Wakeup caused by timer"); break;
    case 4  : Serial.println("Wakeup caused by touchpad"); break;
    case 5  : Serial.println("Wakeup caused by ULP program"); break;
    default : Serial.println("Wakeup was not caused by deep sleep"); break;
  }
}

void vDeepSleepSetup() {
  //Increment boot number and print it every reboot
  ++bootCount;
  Serial.println("Boot number: " + String(bootCount));

  //Print the wakeup reason for ESP32
  print_wakeup_reason();

  /*
    First we configure the wake up source
    We set our ESP32 to wake up every 5 seconds
  */
  esp_sleep_enable_timer_wakeup(TIME_TO_SLEEP * uS_TO_S_FACTOR);
  Serial.println("Setup ESP32 to sleep for every " + String(TIME_TO_SLEEP) +
                 " Seconds");
}

void vGoingToDeepSleep() {
  Serial.print("bootCount: ");
  Serial.println(bootCount);
  Serial.println("Going to sleep now");
  esp_deep_sleep_start();
}
// -----------------------------------------------------------------------------
// } Deep Sleep
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Generic {
// -----------------------------------------------------------------------------
void vDelayESP32(unsigned long ulMilliseconds) {
  unsigned long ulPreviousMillis = millis();

  do {
    yield();
  } while (millis() - ulPreviousMillis <= ulMilliseconds);
}

void vSwap(byte* a, byte* b) {
  byte byAux = *b;

  *b = *a;
  *a = byAux;
}

void float2Bytes(float val, byte* bytes_array) {
  // Create union of shared memory space
  union {
    float float_variable;
    byte temp_array[4];
  } u;
  // Overite bytes of union with float variable
  u.float_variable = val;
  // Assign bytes to input array
  memcpy(bytes_array, u.temp_array, 4);
  vSwap(&bytes_array[0], &bytes_array[3]);
  vSwap(&bytes_array[1], &bytes_array[2]);
}
// -----------------------------------------------------------------------------
// } Generic
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// BME280 {
// -----------------------------------------------------------------------------
void vSetupBME280() {
  // initialize BME280 sensor
  bool status;
  status = bme.begin(0x76);
  if (!status) {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    bBme280isConnected = false;
  } else
    bBme280isConnected = true;
}

void vReadBME280(float* fTc, float* fP, float* fA, float* fRH) {
  *fTc = bme.readTemperature();
  *fP = bme.readPressure() / 100.0F;
  *fA = bme.readAltitude(SEALEVELPRESSURE_HPA);
  *fRH = bme.readHumidity();
}

void vPrintValues(float fT, float fP, float fA, float fH) {
  char sz[35];
  //float fT = bme.readTemperature();
  //float fP = bme.readPressure() / 100.0F;
  //float fA = bme.readAltitude(SEALEVELPRESSURE_HPA);
  //float fH = bme.readHumidity();
  //float fT,fP,fA,fH;
  //vReadBME280(&fT,&fP,&fA,&fH);
  int nY = 0;

  display.clear();

  sprintf(sz, "T: %.2f C", fT);
  display.drawString(0, 0, sz);
  Serial.println(sz);
  sprintf(sz, "P: %.2f hPa", fP);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  sprintf(sz, "A: %.2f m", fA);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  sprintf(sz, "H: %.2f %%", fH);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  Serial.println();

  display.display();
}

void vPrintText(String sz1, String sz2, String sz3, String sz4) {
  char sz[35];
  int nY = 0;

  display.clear();

  sz1.toCharArray(sz, sz1.length()+1);
  display.drawString(0, 0, sz);
  Serial.println(sz);
  sz2.toCharArray(sz, sz2.length()+1);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  sz3.toCharArray(sz, sz3.length()+1);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  sz4.toCharArray(sz, sz4.length()+1);
  display.drawString(0, nY += 15, sz);
  Serial.println(sz);
  Serial.println();

  display.display();
}

void vBme280isNotConnected() {
  display.drawString(0, 15, "BME280");
  display.drawString(0, 30, "is NOT connected");
  Serial.println("BME280 is NOT connected");
}
// -----------------------------------------------------------------------------
// } BME280
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// OLED {
// -----------------------------------------------------------------------------
void vSetupOled() {
  display.init();
  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
}
// -----------------------------------------------------------------------------
// } OLED
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// LMIC {
// -----------------------------------------------------------------------------
void ttnSend(osjob_t* j) {

  // Check if there is not a current TX/RX job running
  if (LMIC.opmode & OP_TXRXPEND) {
    Serial.println(F("[RFM95] Pending message"));
    return;
  }

  // Prepare buffer
  unsigned char data[12];
  int i;
  float fTc, fP, fA, fH;
  //data[0] = (autoincrement >> 24) & 0xFF;
  //data[1] = (autoincrement >> 16) & 0xFF;
  //data[2] = (autoincrement >>  8) & 0xFF;
  //data[3] = (autoincrement >>  0) & 0xFF;
  //float2Bytes(fT,data);
  //void vReadBME280(float* fTc, float* fP, float* fA, float* fRH) {

  vReadBME280(&fTc, &fP, &fA, &fH);
  vPrintValues(fTc, fP, fA, fH);
  if (fTc > -10.f && fTc < 90.) {
    Serial.print("T:");
    Serial.print(fTc);
    Serial.print(" *C <--> ");
    float2Bytes(fTc, data);
    for (i = 0; i < 4 ; i++) {
      Serial.print("[0x");
      Serial.print(data[i], HEX);
      Serial.print("]");
    }
    Serial.println();
    Serial.print("P:");
    Serial.print(fP);
    Serial.print(" hPa <--> ");
    float2Bytes(fP, &data[4]);
    for (i = 4; i < 8 ; i++) {
      Serial.print("[0x");
      Serial.print(data[i], HEX);
      Serial.print("]");
    }
    Serial.println();
    Serial.print("H:");
    Serial.print(fH);
    Serial.print(" % <--> ");
    float2Bytes(fH, &data[8]);
    for (i = 8; i < 12 ; i++) {
      Serial.print("[0x");
      Serial.print(data[i], HEX);
      Serial.print("]");
    }
    Serial.println();

    // Prepare upstream data transmission at the next possible time.
    // Parameters are port, data, length, confirmed
    LMIC_setTxData2(1, data, 12, 0);

    Serial.println(F("[RFM95] Packet queued"));
  }

  // Next TX is scheduled after TX_COMPLETE event.
  autoincrement++;
}

// LMIC library will call this method when an event is fired
void onEvent(ev_t ev) {
  String szRx = "";

  Serial.print(os_getTime());
  Serial.print(": ");

  switch (ev) {
    case EV_SCAN_TIMEOUT:
      Serial.println(F("[RFM95] EV_SCAN_TIMEOUT"));
      break;
    case EV_BEACON_FOUND:
      Serial.println(F("[RFM95] EV_BEACON_FOUND"));
      break;
    case EV_BEACON_MISSED:
      Serial.println(F("[RFM95] EV_BEACON_MISSED"));
      break;
    case EV_BEACON_TRACKED:
      Serial.println(F("[RFM95] EV_BEACON_TRACKED"));
      break;
    case EV_JOINING:
      Serial.println(F("[RFM95] EV_JOINING"));
      break;
    case EV_JOINED:
      Serial.println(F("[RFM95] EV_JOINED"));
      break;
    case EV_RFU1:
      Serial.println(F("[RFM95] EV_RFU1"));
      break;
    case EV_JOIN_FAILED:
      Serial.println(F("[RFM95] EV_JOIN_FAILED"));
      break;
    case EV_REJOIN_FAILED:
      Serial.println(F("[RFM95] EV_REJOIN_FAILED"));
      break;
    case EV_TXCOMPLETE:
      Serial.println(F("[RFM95] EV_TXCOMPLETE (includes waiting for RX windows)"));

      if (TXRX_ACK) {
        Serial.println(F("TXRX_ACK confirmed UP frame was acked\n"));
      }
      Serial.println(F("TX complete ....................................."));
      Serial.print(LMIC.dataLen); Serial.print(" ");
      Serial.println(LMIC.frame[LMIC.dataBeg - 1]);
      for (int i = 0;  i < LMIC.dataLen;  i++) {
        Serial.print(LMIC.frame[LMIC.dataBeg + i], HEX);
        szRx += char(LMIC.frame[LMIC.dataBeg + i]);
      }
      Serial.println(" ");

      if (LMIC.txrxFlags & TXRX_ACK) {
        Serial.println(F("[RFM95] ACK received"));
      }
      if (LMIC.dataLen > 0) {
        Serial.print("Text received: ");
        Serial.println(szRx);
        vPrintText(" ","Rebut: ",szRx," ");
      }
      if (LMIC.dataLen == 1) {
        // Received 1 byte from downlink
        // Content is in LMIC.frame[LMIC.dataBeg]
        /*
          if (LMIC.frame[LMIC.dataBeg] == 0x71){
          Serial.println("Activació de relé");
          PCF_38.write8(0xF7);
          }
          if (LMIC.frame[LMIC.dataBeg] == 0x51){
          Serial.println("Desactivació de relé");
          PCF_38.write8(0xFF);
          }
        */
      }

      if (LMIC.dataLen) {
        Serial.print(F("[RFM95] Received "));
        Serial.print(LMIC.dataLen);
        Serial.println(F(" bytes of payload"));

        for (int i = 0; i < LMIC.dataLen; i++) {
          if (LMIC.frame[LMIC.dataBeg + i] < 0x10) {
            Serial.print(F("0"));
          }
          Serial.print(LMIC.frame[LMIC.dataBeg + i], HEX);
          Serial.print(" ");
        }
        Serial.println();
      }

      // Schedule next transmission
      os_setTimedCallback(&sendjob, os_getTime() + sec2osticks(TX_INTERVAL), ttnSend);
      break;

    case EV_LOST_TSYNC:
      Serial.println(F("[RFM95] EV_LOST_TSYNC"));
      break;
    case EV_RESET:
      Serial.println(F("[RFM95] EV_RESET"));
      break;
    case EV_RXCOMPLETE:
      // data received in ping slot
      Serial.println(F("[RFM95] EV_RXCOMPLETE"));
      break;
    case EV_LINK_DEAD:
      Serial.println(F("[RFM95] EV_LINK_DEAD"));
      break;
    case EV_LINK_ALIVE:
      Serial.println(F("[RFM95] EV_LINK_ALIVE"));
      break;
    default:
      Serial.println(F("[RFM95] Unknown event"));
      break;
  }
}

void vSetupTTN() {
  // SPI interface
#ifdef ESP32
  SPI.begin(SCK_GPIO, MISO_GPIO, MOSI_GPIO, NSS_GPIO);
#endif

  // LMIC init
  os_init();

  // Reset the MAC state. Session and pending data transfers will be discarded.
  LMIC_reset();

  LMIC_setClockError(MAX_CLOCK_ERROR * 1 / 100); // A treure

  // Set static session parameters. Instead of dynamically establishing a session
  // by joining the network, precomputed session parameters are be provided.
  // On AVR, these values are stored in flash and only copied to RAM
  // once. Copy them to a temporary buffer here, LMIC_setSession will
  // copy them into a buffer of its own again.
  uint8_t appskey[sizeof(APPSKEY)];
  uint8_t nwkskey[sizeof(NWKSKEY)];
  memcpy_P(appskey, APPSKEY, sizeof(APPSKEY));
  memcpy_P(nwkskey, NWKSKEY, sizeof(NWKSKEY));
  LMIC_setSession (0x1, DEVADDR, nwkskey, appskey);

  // Set up the channels used by the Things Network, which corresponds
  // to the defaults of most gateways. Without this, only three base
  // channels from the LoRaWAN specification are used, which certainly
  // works, so it is good for debugging, but can overload those
  // frequencies, so be sure to configure the full frequency range of
  // your network here (unless your network autoconfigures them).
  // Setting up channels should happen after LMIC_setSession, as that
  // configures the minimal channel set.
  LMIC_setupChannel(0, 868100000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band

  LMIC_setupChannel(1, 868300000, DR_RANGE_MAP(DR_SF12, DR_SF7B), BAND_CENTI);      // g-band
  LMIC_setupChannel(2, 868500000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  LMIC_setupChannel(3, 867100000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  LMIC_setupChannel(4, 867300000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  LMIC_setupChannel(5, 867500000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  LMIC_setupChannel(6, 867700000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  LMIC_setupChannel(7, 867900000, DR_RANGE_MAP(DR_SF12, DR_SF7),  BAND_CENTI);      // g-band
  LMIC_setupChannel(8, 868800000, DR_RANGE_MAP(DR_FSK,  DR_FSK),  BAND_MILLI);      // g2-band

  // TTN defines an additional channel at 869.525Mhz using SF9 for class B
  // devices' ping slots. LMIC does not have an easy way to define set this
  // frequency and support for class B is spotty and untested, so this
  // frequency is not configured here.

  // Disable link check validation
  LMIC_setLinkCheckMode(0);

  // TTN uses SF9 for its RX2 window.
  LMIC.dn2Dr = DR_SF9;

  // Set data rate and transmit power for uplink (note: txpow seems to be ignored by the library)
  LMIC_setDrTxpow(DR_SF7, 14);

  // Start job
  ttnSend(&sendjob);
}
// -----------------------------------------------------------------------------
// } LMIC
// -----------------------------------------------------------------------------

void setup() {
  pinMode(BUTTON_PRG, INPUT_PULLUP);
  Serial.begin(115200);
  //Wire.begin(I2C_SDA, I2C_SCL);

  vDeepSleepSetup();
  vSetupBME280();
  vSetupOled();
  if (bBme280isConnected)
    vSetupTTN();
  else
    vBme280isNotConnected();

  delayTime = 5000;
}

void loop() {
  static unsigned long ulPreviousMillis = millis();
  unsigned long ulMilliseconds = 5000; // 5 seconds

  os_runloop_once(); // useless using Deep Sleep

  if (millis() - ulPreviousMillis > ulMilliseconds) {
    //vDelayESP32(delayTime); // Time Oled is brighting
    display.clear();
    display.display();
    vGoingToDeepSleep();
  }
}



